//*****************************************************************
//* PCX to Neo Geo FIX converter.
//*****************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>

#define PAL_SIZE		16
#define TILESIZE		64
#define TILEWIDTH		8

typedef unsigned char  BYTE;
typedef unsigned short WORD;
typedef BYTE           BOOL;

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

int  Rip(char *filename);// function declarations                
void WriteSprites(void);
int  LoadPCX(const char* file, BYTE *outpix, BYTE * pal);
long FileSize(const char *fname);
void GetSprite(BYTE *sprite, int x, int y, int width, int height);

FILE   *fp;
FILE   *stream;              // level output data file
WORD    size = 0;
int     Width, Height, pal, start;
WORD    palval;
FILE	*handle;
char	filebank1[32];
char	filepal[32];
BYTE    bank1[65536];
WORD	palette[PAL_SIZE];
long	offset;

BYTE	XTable[]	= {5, 4, 7, 6, 1, 0, 3, 2};

typedef struct PCXHeader
{
    BYTE id;
    BYTE version;
    BYTE rle;
    BYTE bpp;
    WORD xstart;                         
    WORD ystart;
    WORD xend;
    WORD yend;
    WORD hres;
    WORD vres;
    BYTE pal[48];
    BYTE rsvd1;
    BYTE nbitp;
    WORD bytesperline;
    WORD paltype;
    WORD hsize;
    WORD vsize;
    BYTE rsvd2[54];
} PCX;

BYTE PCXBuffer[64000];
BYTE PCXPalette[768];

void Usage();

void main(int argc, char *argv[])
{
    printf("PCX to Neo Geo FIX converter.\n");

    if (argc < 4)
    {
        Usage();
    }

	memset(bank1, 0, sizeof(bank1));
	memset(palette, 0, PAL_SIZE * 2);
    
	offset = atol(argv[1]) + 24576;

    strcpy(filebank1, argv[3]);

    if ("" == filebank1)
    {
        Usage();
    }

	long length;

    handle = fopen(filebank1, "rb");

	if (handle != NULL)
    {
		fseek(handle, 0, SEEK_END);
		length = ftell(handle);
		rewind(handle);

		if (length > sizeof(bank1))
		{
			length = sizeof(bank1);
		}

        fread(bank1, 1, length, handle);
		fclose(handle);
    }

    if (argc > 4)
    {
        strcpy(filepal, argv[4]);

        if (strlen(filepal) <= 0)
        {
            printf("Palette will not be saved.\n");

            goto  EnterWidth;
        }
    }

    else
    {
        memset(filepal, 0, sizeof(filepal));
    }

EnterWidth:
    if (-1 == Rip(argv[2]))
    {
        exit(1);
    }

    WriteSprites();
}

void Usage()
{
    printf("Command syntax is: FixNG Offset Infile Rom [Palette]\n");

    exit(1);
}

//**************************** Rip *******************************
int Rip(char *filename)
{
    LoadPCX(filename, PCXBuffer, PCXPalette);

    if ((Width & 7) != 0 && (Height & 7) != 0)
    {
        printf("FIX character width must be a multiple of 8.\n");
        
		return -1;
    }

    BYTE Tile[TILESIZE];
    BYTE *tile1		= bank1 + offset;

    for (int yloop = 0; yloop < (Height / 8); yloop++)
    {
        for (int xloop = 0; xloop < (Width / 8); xloop++)
        {
            long	x	= xloop * 8;
            long	y	= yloop * 8;

            GetSprite(&Tile[0], x, y, 8, 8);

            for (int tilex = 0; tilex < 8; tilex += 2)
            {
                for (int tiley = 0; tiley < 8; tiley++)
                {

					*tile1	=	((Tile[tiley * TILEWIDTH + XTable[tilex]] & 0x0f) << 4) |
								(Tile[tiley * TILEWIDTH + XTable[tilex + 1]] & 0x0f);
					
					tile1++;
				}
			}
        }
    }


    return 0;
}

void WriteSprites(void)
{
    if (strlen(filepal) > 0)
	{
		for (int palloop = 0; palloop < PAL_SIZE; palloop++)
		{
			BYTE red    = ((PCXPalette[palloop * 3]) >> 2);
			BYTE green  = ((PCXPalette[palloop * 3 + 1]) >> 2);
			BYTE blue   = ((PCXPalette[palloop * 3 + 2]) >> 2);

			red	  &= 0x0f;
			green &= 0x0f;
			blue  &= 0x0f;

			WORD	value = ((WORD)green << 12) | ((WORD)blue << 8) | ((WORD)red << 0);
		    palette[palloop] = value;	
		}

		stream = fopen(filepal, "wb");

		if (NULL == stream)
		{
			return;
		}

		fwrite(palette, 2, sizeof(palette) / sizeof(WORD), stream);
		fclose(stream);
	}

    // open a binary file for the level array
    stream = fopen(filebank1, "wb");

	if (NULL == stream)
	{
		return;
	}

	fwrite(bank1, 1, sizeof(bank1), stream);

    fclose(stream);

	printf("OffSet: %d\n", offset);
}

int LoadPCX(const char* file, BYTE *outpix, BYTE * pal)
{
    int         i, w, h;
    PCXHeader   hdr;
    long        pos;
    int         ret = 0;

    fp = fopen(file, "rb");

    if (fp == NULL)
        goto bye;

    pos = ftell(fp);
    fread (&hdr, 1, sizeof(hdr), fp);

    if (hdr.id != 0x0A || hdr.version != 5 || hdr.rle != 1
     || (hdr.bpp == 8 && hdr.nbitp != 1))
        goto bye;

    w       = hdr.xend - hdr.xstart + 1;
    Width   = w;

    h       = hdr.yend - hdr.ystart + 1;
    Height  = h;

    if (hdr.bpp == 1 && hdr.nbitp == 4)
    {
         memset(pal, 0, 768);

         for (i = 0; i < 48; i++)
            pal[i] = hdr.pal[i] >> 2;
    }

    else if (hdr.bpp == 8 && hdr.nbitp == 1)
    {
        fseek(fp, pos + FileSize(file) - 768, SEEK_SET);

        for (i = 0; i < 768; i++)
            pal[i] = ((BYTE)getc(fp)) >> 2;
    }

    else
        goto bye;

    ret = 1;

    fseek(fp, pos + sizeof(PCXHeader), SEEK_SET);

    while (h-- > 0)
    {
        BYTE   c;
        BYTE  *outpt;
        int             np = 0;

        outpt = outpix;

        memset(outpix, 0, w);

        for (np = 0; np < hdr.nbitp; np++)
        {
            i = 0;
            outpix = outpt;

            do
            {
                c = (BYTE)getc(fp);

                if ((c & 0xC0) != 0xC0)
                {
                    if (hdr.bpp == 1)
                    {
                        int k;
                        for (k = 7; k >= 0; k--)
                            *outpix++ |= ((c >> k) & 1) << np;
                        i += 8;
                    }

                    else
                    {
                        *outpix++ = c;
                        i++;
                    }
                }

                else
                {
                    BYTE v;
                    v = (BYTE)getc(fp);
                    c &= ~0xC0;

                    while (c > 0 && i < w)
                    {
                        if (hdr.bpp == 1)
                        {
                            int k;

                            for (k = 7; k >= 0; k--)
                                *outpix++ |= ((v >> k) & 1) << np;

                            i += 8;
                        }

                        else
                        {
                            *outpix++ = v;
                            i++;
                        }

                        c--;
                    }
                }
            }
            while (i < w);
        }
    }

  bye:
    if (fp != NULL)
        fclose(fp);

    return ret;
}

long FileSize(const char *fname)
{
    FILE *f;
    long length;

    f = fopen(fname, "rb");

    if (f == NULL)
        return 0;

    fseek(f, 0, SEEK_END);

    length = ftell(f);

    fclose(f);

    return length;
}

void GetSprite(BYTE *tile, int x, int y, int width, int height)
{
    BYTE *localtile = tile;
    BYTE *buffer    = &PCXBuffer[0];

    buffer += y * Width + x;

    for (int outerloop = 0; outerloop < height; outerloop++)
    {
        for (int innerloop = 0; innerloop < width; innerloop++)
        {
            *localtile++ = *buffer++;
        }

        buffer += Width - width;
    }
}
